#!/bin/bash
# $File: /usr/local/share/smtpauth2ban/lib/function.inc.sh $
# $Author: Marko Schulz <info@tuxnet24.de> $
# $Date: 2023-11-16 16:08:14 +0200 (Do, 16 Nov 2023) $
# $HeadURL: https://gitbucket.tuxnet24.de/git/tuxnet24/smtpauth2ban.git $
# $Description: Function library of the smtpauth2ban script. $
# *******************************************************
#
# [GLOBAL FUNCTIONS]
#
# *******************************************************
# This function is like the php function trim,
# witch removes all spaces in front & back of a string.
#
# @param <STDIN> - The string as pipe stream
# @return string
#
function trim () {

cat - | perl -p -e 's/^\s+|\s+$//g'

}

# *******************************************************
# This function converts time designations
# such as 1d or 12h into seconds.
#
# @param $1 - The time designations
# @return init
#
function datestr2seconds () {

local item=$1
local digit=$(echo $item | perl -n -e 'print $1," ",$2 if ($_ =~ m/^(\d+)((?:d|h|m|s))$/g);' | awk '{print $1}')
local unit=$(echo $item | perl -n -e 'print $1," ",$2 if ($_ =~ m/^(\d+)((?:d|h|m|s))$/g);' | awk '{print $2}')
local sec

case $unit in
        d) sec=$((${digit}*86400)) ;;
        h) sec=$((${digit}*3600))  ;;
        m) sec=$((${digit}*60))    ;;
        s) sec=${digit}            ;;
esac

echo $sec

}

# *******************************************************
# This function check, if the item is an ip address.
#
# @param $1 - ip address
# @return bool
#
function is_ipaddr () {

if [[ "$1" =~ (([01]{,1}[0-9]{1,2}|2[0-4][0-9]|25[0-5])\.([01]{,1}[0-9]{1,2}|2[0-4][0-9]|25[0-5])\.([01]{,1}[0-9]{1,2}|2[0-4][0-9]|25[0-5])\.([01]{,1}[0-9]{1,2}|2[0-4][0-9]|25[0-5]))$ ]]; then
    return 0
else
    return 1
fi

}

# *******************************************************
# This function get a string and write them as log message
# to the logfile ${cfg[logfile]} (append mode).
#
# @param $* - The log message
# @return void
#
function logger () {

echo -e "[$(date +%Y-%m-%d\ %H:%M:%S)] $*" >>${cfg[logfile]}

}

# *******************************************************
# This function add or delete (DROP) an defined ip address
# from the defined firewall chain.
#
# @param $1 - mode (add|del)
# @param $2 - ip address
# @param $3 - rule chain
# @return void
#
function firewall () {

if [ ! -n "$2" -a ! -n "$3" ]; then
    return 1
else
    case $1 in
        add)
            return $( fw_ban $3 $2 )
        ;;
        del)
            return $( fw_unban $3 $2 )
        ;;
        *)
            return 2
        ;;
    esac
fi

}

# *******************************************************
# This function check with iptables, if a defined chain exists.
#
# @param $1 - fw chain name
# @return bool
#
function fw_chain_exists () {

[ ! -n "$1" ] && return 1
iptables -n --list "$1" >/dev/null 2>&1
return $?

}

# *******************************************************
# This function create or delete a defined fw chain.
#
# @param $1 - fw mode (add|del)
# @param $2 - fw chain name
# @return bool
#
function fw_chain () {

# Return failure, if first parameter is no add or del
if [ "$1" != "add" -a "$1" != "del" ]; then
    return 1
fi

# Return failure, if no chain name was defined
if [ ! -n "$2" ]; then
    return 2
fi

# Now we can add or del a defined chain
case $1 in
    add)
        if fw_chain_exists "${2}" ; then
            return 0
        else
            # Create new chain
            iptables -N $2
            # APPEND a RETURN command to the end of the chain (same as return in other languages)
            iptables -A $2 -j RETURN
            # INSERT a rule at the start of the built-in INPUT chain that redirects all packets to the chain
            iptables -I INPUT -p tcp -j $2
            return $?
        fi
    ;;
    del)
        if fw_chain_exists "${2}" ; then
            # Delete chain ($2) from INPUT chain (de-reference)
            iptables -D INPUT -j $2
            # Flush (empty) rule on chain ($2)
            iptables -F $2
            # Delete chain ($2)
            iptables -X $2
            return $?
        else
            return 0
        fi
    ;;
    *)
        return 3
    ;;
esac

}

# *******************************************************
# This function add a defined ip address to a defined fw chain.
#
# @param $1 - fw chain name
# @param $2 - ip address
# @return bool
#
function fw_ban () {

iptables -I $1 1 -s $2 -j DROP
return $?

}

# *******************************************************
# This function remove a defined ip address from a defined fw chain.
#
# @param $1 - fw chain name
# @param $2 - ip address
# @return bool
#
function fw_unban () {

iptables -D $1 -s $2 -j DROP
return $?

}

# *******************************************************
# This function updates the firewall depending on the cfg[bannedfile].
#
# @return void
#
function run_banning () {

# Create these files, if not exists
[ ! -f "${cfg[bannedfile]}" ] && touch ${cfg[bannedfile]}
[ ! -f "${cfg[whitelist]}" ]  && touch ${cfg[whitelist]}
[ ! -f "${newaddress}" ]      && touch ${newaddress}

# Get possibily new remote addresses that will be banned
local newips=$(
    journalctl --since "$(date --date=${cfg[minutes_before]}' minutes ago' +%Y-%m-%d\ %H:%M:%S)" -u ${cfg[service]} -p ${cfg[priority]} --no-pager | \
        perl -lne 'print $1 if $_ =~ m/warning\:\s+unknown\[([^\]]+)\]\:\s+SASL LOGIN authentication failed\:/g;' | \
        sort -u | \
        grep -v -f ${cfg[whitelist]}
)

# Create a firewall chain, to ban ip addresses
if [ "${cfg[detection]}" = "false" ]; then
    # Add the firewall chain to the iptables INPUT chain
    if ! fw_chain "add" "${cfg[firewall_chain]}"; then
        echo -e "ERROR: Can't create the firewall chain ${cfg[firewall_chain]}." && exit 1
    fi
fi

# Create a temporary file with the new addresses
echo -e "$newips" >${newaddress}

# Get the current timestamp
local ctime=$(date +%s)

local toadd
# New addresses to add with difference from whitelist
for toadd in $( grep -v -f ${cfg[whitelist]} ${newaddress} ); do
    toadd=$( echo $toadd | trim )
    # Check, if $item already exist in the firewall
    if ! grep -q ${toadd} ${cfg[bannedfile]}; then
        if is_ipaddr "${toadd}" ; then
            # Temporary variable, if an error occurred
            local failure=0
            # Add remote address to iptables rules into the firewall chain
            if [ "${cfg[detection]}" = "true" ]; then
                logger "${cfg[hostname]}" "BAN" "${toadd}"
            else
                if ! firewall "add" "${toadd}" "${cfg[firewall_chain]}"; then
                    failure=1
                    echo -e "ERROR: Can't ban the ip address ${toadd}!" && exit 1
                else
                    logger "${cfg[hostname]}" "BAN" "${toadd}"
                fi
            fi
            if [ $failure -eq 0 ]; then
                # Add the new ip address as banned to the banned file
                echo -ne "${ctime}|${toadd}\n" >>${cfg[bannedfile]}
                # Add the new ip address as banned to the firewall chain
                if  [[ $TERM =~ ^xterm ]]; then
                    echo -e "\e[41m\e[1;37m fw add \e[0m ${toadd}"
                fi
            fi
        fi
    fi
done

local d=0
local todel
# Delete expiered ip addresses
for todel in $( cat ${cfg[bannedfile]} ); do
    # Get timestamp and ip address
    local ts=$( echo $todel | awk -F'|' '{print $1}' )
    local ip=$( echo $todel | awk -F'|' '{print $2}' | trim )
    if is_ipaddr "${ip}" ; then
        # Temporary variable, if an error occurred
        local failure=0
        # Calculate the expire timestamp
        local expts=$[$ts+$( datestr2seconds ${cfg[banntime]} )]
        # Remove the ip address from firewall, if the banned time has expiered
        if [ $expts -lt $ctime ]; then
            # Del remote address to iptables rules into the firewall chain
            if [ "${cfg[detection]}" = "true" ]; then
                logger "${cfg[hostname]}" "UNBAN" "${ip}"
            else
                if ! firewall "del" "${ip}" "${cfg[firewall_chain]}"; then
                    failure=1
                    echo -e "ERROR: Can't unban the ip address ${ip}!" && exit 1
                else
                    logger "${cfg[hostname]}" "UNBAN" "${ip}"
                fi
            fi
            if [ $failure -eq 0 ]; then
                # firewall "del" ${ip}
                if  [[ $TERM =~ ^xterm ]]; then
                    echo -e "\e[1;43m\e[1;37m fw del \e[0m ${ip}"
                fi
                d=1
            fi
        else
            # Write bannfile with no unexpired ip addresses
            echo -ne "${ts}|${ip}\n" >>${cfg[bannedfile]}.new
        fi
    fi
done

# Create the bannedfile
if [ -f "${cfg[bannedfile]}.new" ]; then
    mv ${cfg[bannedfile]}.new ${cfg[bannedfile]}
else
    if [ $d -eq 1 ]; then
        rm ${cfg[bannedfile]}
        touch ${cfg[bannedfile]}
    fi
fi

# clean temporary file
rm ${newaddress}

}

# *******************************************************
# This is the main function. Ther will be som variables
# prepared and then run the run_banning() method.
#
# @return void
#
function main () {

# Replace the placeholder in the path of cfg[newaddress]
newaddress=$( printf ${cfg[newaddress]} "${cfg[hostname]}" )

# Run the banning method
run_banning

if  [[ $TERM =~ ^xterm ]]; then
        echo -e "\e[42m\e[1;37m  next  \e[0m"
fi

}

# vim: syntax=bash ts=2 sw=2 sts=2 sr noet
# EOF
